// hwvTimeTable_DB.cpp Time-stamp: <2006/03/24 21:43 haru@XP4PDA05C>
// (C) 2005-2006 HAL <http://hal9.blog3.fc2.com> HAL.Kinosita@gmail.com

/*
 * TimeTable Viewer
 * Copyright(c)1999 by Hiroaki Sengoku <sengoku@gcd.org>
 * Version 0.1	Jun 19, 1999
 * Version 0.2	Jun 20, 1999
 * Version 0.3	Jun 20, 1999
 * Version 0.4	Jun 21, 1999
 * Version 0.5	Jun 22, 1999	AboutDialog
 * Version 0.6	Jun 25, 1999	SelectTrain, HardButton, PrefsDialog
 * Version 0.7	Jun 26, 1999	scroll bug fix
 * Version 0.8	Jul  3, 1999	Arrive, ShareReg, History, Transfer, StopClock
 * Version 0.9	Jul  8,	1999	multi '>', ContFile, CommentMark, scramble
 * Version 1.0	Jul 31, 1999	split bar, DB, ;;CANCEL, '\' esc, HeadType, LRU
 * Version 1.1	Aug 26, 1999	checkTimeOut
 * Version 1.2	Aug 27, 2000	Alarm, InstalledSec, LastUsedSec
 * Version 1.3  Aug  1, 2001	GPL version
 */

#include <PalmOS.h>
#include <StdIOPalm.h>
#define HwvTimeTable_DB 1
#include "hwvTimeTable.h"
#undef HwvTimeTable_DB

//========== DB.c ================================================================================================
void initializeDB(void) {
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	MemHandle appInfoH;
	gTTP->MemoDB = DmOpenDatabaseByTypeCreator('DATA', 'memo', dmModeReadOnly);
	if(gTTP->MemoDB){
		gTTP->TrainTimeCategory = CategoryFind(gTTP->MemoDB, "TrainTime");
		appInfoH = getAppInfo(gTTP->MemoDB);
		if(appInfoH){
	    MemoAppInfoType *appInfoP = (MemoAppInfoType*)MemHandleLock(appInfoH);
	    if(appInfoP->sortOrder) 
				gTTP->InsertPosMemo = InsertPosAlphabetical;
	    MemHandleUnlock(appInfoH);
		}
	}
	else{
		; //PrivErrSetCode(DmGetLastErr());
		//PrivErrDisplay("Can't open MemoPad",true);
	}
}

static Int16 isConfigFile(UInt16 recordNum){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	Int16 type = 0;	/* assume not config file */
	MemHandle h = DmQueryNextInCategory(gTTP->MemoDB, &recordNum, dmAllCategories);
	Char *p;
	if(!h) return -1;	/* not exist */
	p = (Char*)MemHandleLock(h);
	if(      StrNCompare(p,"TrainTime",        9) == 0 && p[9 ] == linefeedChr){
		type = 1;	/* TrainTime ConfigFile */
	}else if(StrNCompare(p,"TimeTable Viewer",16) == 0 && p[16] == linefeedChr){
		type = 2;	/* TimeTable Viewer ConfigFile */
	}
	MemHandleUnlock(h);
	gTTP->ConfigHandle = h;
	return type;
}

static void configFile(void){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	UInt16 trainNum = 0;
	MemHandle trainH = NULL;
	UInt16 recordNum = 0;
	Int16 type;
	while((type=isConfigFile(recordNum)) >= 0){
		switch (type){
			case 1:
				trainNum = recordNum;
				trainH = gTTP->ConfigHandle;
				break;
			case 2:
				gTTP->ConfigRecordNum = recordNum;
				gTTP->ConfigHandle = gTTP->ConfigHandle;
				return;
		}
		recordNum++;
	}
	if(trainH){
		gTTP->ConfigRecordNum = trainNum;
		gTTP->ConfigHandle = trainH;
		return;
	}
	gTTP->ConfigRecordNum = -1;
	gTTP->ConfigHandle = NULL;
}

void initializeDBAfterPrefs(void){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	if(gTTP->ConfigRecordNum < 0 || isConfigFile(gTTP->ConfigRecordNum) <= 0)
		configFile();
}

TimeTableType *newTimeTable(void){
	TimeTableType *table = (TimeTableType*)MemPtrNew(sizeof(TimeTableType));
	MemPtrSetOwner(table, 0);

	if(!table){
		WinDrawLine(80, 80, 120, 0);
		return NULL;
	}
	table->recordNum = -1;
	table->Lines = NULL;
	table->LinesLen = 0;
	table->LinesBuf = NULL;
	table->Comments = NULL;
	table->CommentsBuf = NULL;
	table->Arrives = NULL;
	table->ArrivesLen = 0;
	table->ArrivesBuf = NULL;
	table->NarrowsLen = 0;
	table->Narrows = NULL;
	return table;
}
void disposeTimeTable(TimeTableType *table){
	if(table->Lines) MemPtrFree(table->Lines);
	if(table->LinesBuf) MemPtrFree(table->LinesBuf);
	if(table->Comments) MemPtrFree(table->Comments);
	if(table->CommentsBuf) MemPtrFree(table->CommentsBuf);
	if(table->Arrives) MemPtrFree(table->Arrives);
	if(table->ArrivesBuf) MemPtrFree(table->ArrivesBuf);
	if(table->Narrows) MemPtrFree(table->Narrows);
	MemPtrFree(table);
}

void skipSpace(Char **pp){
	Char *p;
	p = *pp;
	while(*p == ' ' || *p == '\t') p++;
	*pp = p;
}

static void skipLine(Char **pp){
	Char *p, *q;
	p = *pp;
	q = StrChr(p, linefeedChr);
	if(q){ while(*q == linefeedChr) q++; }
	else {	q = p + StrLen(p);           }
	*pp = q;
}

static Int8 conv2digit(Char **pp){
	Char *p;
	Int16 i;
	p = *pp;
	if('0' <= *p && *p <= '9'){
		i = *p - '0';
		if('0' <= *(p+1) && *(p+1) <= '9'){
	    i *= 10;
	    p++;
	    i += *p - '0';
		}
		p++;
		*pp = p;
	}else{
		i = -1;	/* error */
	}
	return i;
}

static Int16 dateToken(Char **pp){
	Int16 mon, day;
	skipSpace(pp);
	mon = (Int16)conv2digit(pp);
	skipSpace(pp);
	if(mon < 0 || **pp != '/') return -1;
	(*pp)++;
	skipSpace(pp);
	day = (Int16)conv2digit(pp);
	skipSpace(pp);
	if(day < 0) return -1;
	return day + (mon << 8);
}

static WeekType weekTypeToken(Char **pp){
	const Char WeekChars[] =
		"ALL ORD MON TUE WED THU FRI SAT SUN HOL "
		"UR0 UR1 UR2 UR3 UR4 UR5 UR6 UR7 UR8 UR9 ";
	Char str[4];
	Char *p, *q;
	Int16 i = 0;
	WeekType ret = Errday;
	p = *pp;
	while(*p != '[' && *p != '*' && *p && *p != linefeedChr){
		p++;
	}
	if(*p == '[' || *p == '*'){
		p++;
	}else{
		*pp = p;
		return ret;
	}
	while(*p != ']' && *p != '*' && *p != linefeedChr && *p){
		if(i < 3) str[i++] = *p;
		p++;
	}
	if(*p == ']') p++;
	str[i] = '\0';
	*pp = p;
	q = StrStr(WeekChars,str);
	if(q){
		ret = (WeekType)((q - WeekChars)/4 + Allday);
	}
	return ret;
}

static WeekType parseConfig(MemHandle h){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	WeekType week = Errday;
	Int16 from, end, today;
	//byHAL//DateTimeType dateTime;
	DateTimeType dateTime = gTTP->gHWP->cDateTime;
	//byHAL//UInt32 now = TimGetSeconds();
	UInt32 now = TimDateTimeToSeconds(&(gTTP->gHWP->cDateTime)); //byHAL
	Char *p;
	if(!h) return week;

	//TimSecondsToDateTime(now, &dateTime);
	if(dateTime.hour <= MidNight){
		now -= ((dateTime.hour+1) * 60L * 60L);	/* yesterday */
		TimSecondsToDateTime(now, &dateTime);
	}
	today = dateTime.day + (dateTime.month << 8);
	p = (Char*)MemHandleLock(h);
	skipLine(&p);
	while(*p){
		skipSpace(&p);
		if(*p == '['){
	    week = weekTypeToken(&p);
	    skipLine(&p);
		}else if(*p == ';'){
	    skipLine(&p);
		}else if('0' <= *p && *p <= '9'){
	    from = dateToken(&p);
	    if(*p == '-'){
				p++;
				end = dateToken(&p);
	    }else{
				end = from;
	    }
	    if(0 <= from && from <= today
				 && 0 <= end && today <= end) goto exit;
	    skipLine(&p);
		}else{
	    skipLine(&p);
		}
	}
	week = Errday;
 exit:
	MemHandleUnlock(h);
	return week;
}

static Int16 convDigits(Char **pp){
	Char *p;
	Int16 i = 0;
	Boolean err = true;
	p = *pp;
	while('0' <= *p && *p <= '9'){
		err = false;
		i *= 10;
		i += *p - '0';
		p++;
	}
	if(err){
		return -1;
	}else{
		*pp = p;
		return i;
	}
}

/* length of the first line and *pp points the next line */
Int16 getLine(Char **pp){
	Char *p, *q;
	Int16 len;
	p = *pp;
	q = StrChr(p,linefeedChr);
	if(q){
		len = q - p;
		q++;
	}else{
		len = StrLen(p);
		q = p + len;
	}
	*pp = q;
	return len;
}

Boolean isContFile(Char **pp){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	Char *p, *q, *r;
	Boolean ret = false;
	Int16 len, n, i;
	q = *pp;
	len = getLine(&q);
	p = *pp;
	for (i=len-1; i > 0; i--){
		if(p[i] == ' '){
	    r = &p[i+1];
	    n = convDigits(&r);
	    if(i > STRMAX-1) i = STRMAX - 1;
	    if(gTTP->ContFileNo <= 0 && n == 1){			/* 1st file */
				StrNCopy(gTTP->ContFileName,p,i);
				gTTP->ContFileName[i] = '\0';
				if(StrNCompare(q,gTTP->ContFileName,i) == 0){	/* 2nd line */
					gTTP->ContFileNo = n;
					ret = true;
				}
	    }else if(0 < gTTP->ContFileNo && n == gTTP->ContFileNo + 1
								&& StrNCompare(p,gTTP->ContFileName,i) == 0	/* 2nd file */
								&& gTTP->ContFileName[i] == '\0'){
				gTTP->ContFileNo = n;
				ret = true;
	    }
	    break;
		}
		if(p[i] < '0' || '9' < p[i]) break;	/* not digit */
	}
	if(ret){
		*pp = q;
	}else{
		gTTP->ContFileNo = -1;
		gTTP->ContFileName[0] = '\0';
	}
	return ret;
}

static Char contNextChar(Char **pp){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	Char *p;
	Int16 index, n;
	p = *pp;
	if(gTTP->ContFileNo == 0){	/* initialize */
		gTTP->NContFiles = 0;
		gTTP->ContFileHandle = NULL;
		gTTP->ContFileLastLine = NULL;
		index = 0;
		gTTP->ContFileLastLineBuf[index] = '\0';
	}else{
		if(gTTP->ContFileNextLine){
	    if(*p == '\0'){	/* get from ContFileLastLineBuf */
				p = *pp = gTTP->ContFileNextLine;
				gTTP->ContFileNextLine = NULL;
	    }
	    if(*p) return *p;
		}
		if(gTTP->ContFileNo < 0){	/* no more continue */
	    if(*p == '\0'){
				if(gTTP->ContFileHandle) MemHandleUnlock(gTTP->ContFileHandle);
				gTTP->ContFileHandle = NULL;
				return '\0';
	    }
	    return *p;
		}
		if(p < gTTP->ContFileLastLine) return *p;
		for (index=0; index < STRLINEMAX-1 && *p; index++)
	    gTTP->ContFileLastLineBuf[index] = *p++;
		gTTP->ContFileLastLineBuf[index] = '\0';
	}
	if(gTTP->ContFileHandle) MemHandleUnlock(gTTP->ContFileHandle);
	gTTP->ContFileNextLine = NULL;
	if(gTTP->ContFileRecNo >= TimeTableRecordNumMin && gTTP->TimeTableDB){
		UInt16 recordNum = gTTP->ContFileRecNo - TimeTableRecordNumMin;
		gTTP->ContFileHandle = DmQueryNextInCategory(gTTP->TimeTableDB, &recordNum, gTTP->TimeTableCategory);
		gTTP->ContFileRecNo = recordNum + TimeTableRecordNumMin;
	}else{
		UInt16 recordNum = gTTP->ContFileRecNo;
		gTTP->ContFileHandle = DmQueryNextInCategory(gTTP->MemoDB, &recordNum, gTTP->TrainTimeCategory);
		gTTP->ContFileRecNo = recordNum;
	}
	gTTP->ContFileRecNo++;
	n = gTTP->ContFileNo;
	if(gTTP->ContFileHandle){
		p = (Char*)MemHandleLock(gTTP->ContFileHandle);
		gTTP->ContFileLastLine = p + MemHandleSize(gTTP->ContFileHandle) - 1;
		while(*(gTTP->ContFileLastLine) != linefeedChr) gTTP->ContFileLastLine--;
		gTTP->ContFileLastLine++;
		if(isContFile(&p)){
	    gTTP->NContFiles++;
	    for( ; index < STRLINEMAX-1 && *p && *p != linefeedChr; index++)
				gTTP->ContFileLastLineBuf[index] = *p++;
	    if(*p == linefeedChr) gTTP->ContFileLastLineBuf[index++] = *p++;
	    gTTP->ContFileLastLineBuf[index] = '\0';
	    if(*p == linefeedChr) p++;
	    gTTP->ContFileNextLine = p;
	    p = *pp = gTTP->ContFileLastLineBuf;
	    return *p;
		}else if(n <= 0){	/* not continuous file */
	    gTTP->NContFiles = 1;
	    *pp = p;
	    return *p;
		}else{		/* end of continuous file */
	    MemHandleUnlock(gTTP->ContFileHandle);
	    gTTP->ContFileHandle = NULL;
		}
	}
	gTTP->ContFileNo = -1;
	p = *pp = gTTP->ContFileLastLineBuf;
	return *p;
}

/* begin with '#' */
void titleLine(Char **pp){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	Int16 length;
	Char *p, *q;
	p = *pp;
	if(*p == '#' || *p == ';') p++;
	skipSpace(&p);
	q = p;
	length = getLine(&q);
	if(length >= STRMAX - 1) length = STRMAX - 1;
	StrNCopy(gTTP->Title,p,length);
	gTTP->Title[length] = '\0';
	*pp = q;
}

/* begin with '[' */
static Boolean weekTypeLine(Char **pp){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	WeekType week;
	Boolean ret = false;
	Boolean comb;
	Char *p;
	p = *pp;
	while(*p && *p != linefeedChr){
		comb = (*p == '*');
		if(!comb && ret) break;
		week = weekTypeToken(&p);
		if(week == Errday){
	    continue;
		}else if(week == Allday){
	    ret = true;
		}else if(week == Weekday){
	    if((Monday <= gTTP->Today && gTTP->Today <= Friday)
				 || gTTP->Today == Weekday) ret = true;
		}else if(week == gTTP->TodayDate){
	    ret = true;
		}else if(gTTP->Today == Weekday){
	    if((Monday <= week && week <= Friday)
				 || week == Weekday) ret = true;
		}else if(gTTP->Today == week){
	    ret = true;
		}else if(comb){
	    ret = false;	/* conbination with previous conditions */
		}
	}
	skipLine(&p);
	*pp = p;
	return ret;
}

/* convert binary string to integer */
static Int32 btoi(Char **pp){
	Int32 ret = 0;
	Char *p;
	p = *pp;
	if(*p != '0' && *p != '1') return -1;
	do {
		ret <<= 1;
		ret |= (*p - '0');
		p++;
	} while(*p == '0' || *p == '1');
	*pp = p;
	return ret;
}

/* begin with ';;' */
static void directiveLine(Char **pp){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	Char *p;
	Int16 len, n;
	Int32 dot;
	p = *pp;
	len = getLine(pp);
	if(len <= 0) return;
	if(     StrNCompare(p,"UNMARK"  ,len) == 0) gTTP->CommentMark = UnMark;
	else if(StrNCompare(p,"SUCC"    ,len) == 0) gTTP->CommentMark = SuccMark;
	else if(StrNCompare(p,"PREC"    ,len) == 0) gTTP->CommentMark = PrecMark;
	else if(StrNCompare(p,"MIDDLE"  ,len) == 0) gTTP->CommentMark = MiddleMark;
	else if(StrNCompare(p,"MARK"    ,len) == 0) gTTP->CommentMark = Mark;
	else if(StrNCompare(p,"DOTMARK" ,len) == 0) gTTP->CommentMark = DotMark;
	else if(StrNCompare(p,"DDOTMARK",len) == 0) gTTP->CommentMark = DDotMark;
	else if(StrNCompare(p,"CANCEL"  ,len) == 0) gTTP->CommentMark = CancelMark;
	else if(StrNCompare(p,"MARK"    ,  4) == 0
					 && (p += 4, (n=btoi(&p)) >= 0) && n <= 0x1F){
		gTTP->CommentMark = (CommentMarkType)(Mark00000 + n);
		if(*p++ == '=' && (dot=btoi(&p)) >= 0){
	    gTTP->DotPattern[n] = (UInt16)dot;
		}
	}
}

static Int16 commentChar(Char c){
	Int16 ret;
	if(     'A' <= c && c <= 'Z'){ ret = c - 'A';      }
	else if('a' <= c && c <= 'z'){ ret = c - 'a' + 26; }
	else if(c == '-'            ){ ret = NoComment;    } /* no comment */
	else                         { ret = -1;           }/* not comment char */
	return ret;
}

/* begin with '[A-Za-z]:' */
static Int16 commentLine(Char **pp, Char *buf, Char **comments,
												 CommentMarkType *commentsMark,
												 Char **narrows, Int16 *narrowsLenP){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	Int16 length = 0;
	Int16 comment;
	Char *p, *q;
	Int8 comms[NComments+1];
	Int16 ncomms = 0;
	Int16 i;
	p = *pp;
	while((comment=commentChar(*p)) >= 0){
		if(ncomms < NComments) comms[ncomms++] = comment;
		p++;
	}
	if(ncomms > 0 && *p++ == ':'){
		comment = comms[0];
		q = p;
		while(*q){
	    if(*q == linefeedChr || *q == ';') break;
	    q++;
		}
		length = q - p;
		if(*q) skipLine(&q);
		if(buf){
	    if(comments) comments[comment] = buf;
	    if(narrows && narrowsLenP) narrows[*narrowsLenP] = buf;
	    StrNCopy(buf,p,length);
	    buf[length] = '\0';
		}
		if(narrowsLenP){
	    if((*narrowsLenP) == gTTP->NarrowMode){
				for (i=0; i < ncomms; i++){
					gTTP->NarrowList[i] = comms[i];
				}
				gTTP->NarrowList[i] = -1;	/* delimitter */
	    }
	    (*narrowsLenP)++;
		}
		length++;	/* with delimitter */
		if(buf){
	    for (i=0; i < ncomms; i++){
				buf[length+i] = comms[i];
	    }
	    buf[length+ncomms] = -1;	/* delimitter */
		}
		length += (ncomms + 1);
		if(commentsMark) commentsMark[comment] = gTTP->CommentMark;
		*pp = q;
		return length;
	}else{
		skipLine(pp);
		return 0;
	}
}

static Boolean ishidden(Int8 com, Int8 *comments){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	Int8 *p, *q;
	if(gTTP->NarrowMode <= 0) return false;	/* all */
	if(gTTP->NarrowMode == 1){
		if(com >= 0) return true;	/* without notes */
		return false;
	}
	p = gTTP->NarrowList;
	while(*p >= 0){
		if(*p == NoComment && com < 0) return false;
		if(com == *p) return false;
		q = comments;
		while(*q >= 0){
	    if(*q == *p) return false;
	    q++;
		}
		p++;
	}
	return true;
}

static Boolean convItem(Char **pp, ItemType *item,
												Int8 *comments, Int16 *cip,
												CommentMarkType *commentsMark){
	Char *p;
	Int16 min;
	Int16 com, c;
	CommentMarkType mark;
	Int16 ci;
	p = *pp;
 again:
	mark = UnMark;
	ci = *cip;
	skipSpace(&p);
	com = commentChar(*p);
	if(com >= 0){
		if(mark < commentsMark[com]) mark = commentsMark[com];
		p++;
		while((c=commentChar(*p)) >= 0){
			if(commentsMark[c] >= Mark00000 && mark >= Mark00000){
				mark = (CommentMarkType)(((commentsMark[c] - Mark00000) | (mark - Mark00000)) + Mark00000);
			}else if(mark < commentsMark[c]) mark = commentsMark[c];
			if(ci < ExCommentsMax-1) comments[ci++] = c;
			p++;
		} 
	}
	min = conv2digit(&p);
	if(min >= 0){
		comments[ci++] = -1;	/* separater */
		if(ishidden(com,comments+*cip)) goto again;
		if(item){
			item->comment = com;
			item->minute = min;
			item->mark = (UInt8)mark;
		}
	}
	skipSpace(&p);
	*pp = p;
	if(min >= 0){
		*cip = ci;
		return true;
	}
	return false;
}

static void storeWord(UInt16 w, UInt8 *p){	/* *p = w */
	p[0] = (w >> 8);	/* MSB big endian */
	p[1] = (w & 0xff);	/* LSB */
}

static Int16 convLine(Char **pp, LineType *line,
											Int16 *cip, CommentMarkType *commentsMark){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	static Int8 comments[ExCommentsMax];	/* static to save stack area */
	Int8 *cp;
	Int16 n = 0;
	Int16 i;
	*cip = 0;
	while(**pp && **pp != linefeedChr && n < gTTP->NItems){
		if(convItem(pp,(line ? &line->items[n] : NULL),
								comments, cip, commentsMark)) n++;
		else {
	    skipLine(pp);
	    break;
		}
	}
	while(**pp == linefeedChr) (*pp)++;
	if(*cip == n) *cip = 0;	/* no extra comments */
	if(line){
		storeWord((UInt16)*cip, &line->excom_msb);	/* line->excom = *cip; */
		cp = (Int8*)&line->items[n];
		for (i=0; i < *cip; i++) cp[i] = comments[i];
	}
	return n;
}

/* begin with [0-9]+: */
static Int16 hourLine(Char **pp, LineType *line, LineType **lines, Int16 *lenP,
											CommentMarkType *commentsMark){
	Int16 length = 0;
	Int16 hour;
	Int16 excom;
	Int16 i, j;
	Char *p;
	p = *pp;
	if(skipSpace(&p), (hour = conv2digit(&p)) >= 0 && *p++ == ':'){
	multi:
		i = convLine(&p, line, &excom, commentsMark);
		j = sizeof(LineType) + sizeof(ItemType) * i + sizeof(Int8) * excom;
		length += j;
		if(line){
	    line->hour = hour;
	    line->length = i;
	    if(lines && lenP) lines[*lenP] = line;
	    (UInt8*)line += j;
		}
		if(lenP) (*lenP)++;
		if(*p && *(p-1) != linefeedChr) goto multi;	/* multi line */
		*pp = p;
	}else{
		skipLine(pp);
	}
	return length;
}

/* begin with '$' */
static Int16 arriveLine(Char **pp, Char *buf, Char **arrives, Int16 *lenP){
	Int16 stn = 0;
	Int16 length = 0;
	Char *p;
	Int16 comment;
	Int16 min;
	Boolean eod = false;
	Char trans[STRLINEMAX];
	Int16 trlen = 0;
	Int16 trpos = 0;
	Int16 i;
	Boolean kanji;
	p = *pp;
	p++;
	while(*p && *p != ':'){
		if(buf) buf[length] = *p;
		length++;
		p++;
	}
	if(buf){
		if(arrives && lenP) arrives[*lenP] = buf;
		buf[length] = '\0';
	}
	if(lenP){
		stn = *lenP;
		(*lenP)++;
	}
	length++;
	trlen = 0;
	trpos = length;
	length += 2;
	if(buf){
		buf[trpos] = 0;
		buf[trpos+1] = 0;
	}
	skipSpace(&p);
	while(*p && *p != linefeedChr){
		if(*p == ';'){
	    skipLine(&p);
	    break;
		}
		if(*p == '>'){	/* transfer */
	    p++;
	    skipSpace(&p);
	    kanji = false;
	    while(*p && *p != linefeedChr && *p != ' '
						&& *p != '+' && trlen < STRLINEMAX-2){
				if(!kanji && *p == '\\') p++;	/* escape char */
				kanji = isKanji(*p);
				trans[trlen++] = *p++;
	    }
	    trans[trlen++] = '\0';
	    skipSpace(&p);
	    if(*p == '+'){
				p++;
				skipSpace(&p);
				trans[trlen++] = convDigits(&p);
				skipSpace(&p);
	    }else{
				trans[trlen++] = 0;
	    }
	    continue;
		}
		eod = false;
		comment = commentChar(*p);
		if(comment < 0){
	    if(('0' <= *p && *p <= '9') || *p == '='){
				if(buf) buf[length] = (Char)-1;
				length++;
				if(*p == '='){
					min = -2;
					p++;
				}else{
					min = convDigits(&p);
				}
				if(buf) storeWord((UInt16)min,(UInt8*)&buf[length]);
				length += 2;
				eod = true;
	    }else{
				p++;
	    }
	    skipSpace(&p);
	    continue;
		}
		if(buf) buf[length] = (Char)comment;
		length++;
		p++;
	}
	if(!eod){
		if(buf){
	    min = -1;
	    buf[length] = (Char)-1;
	    storeWord((UInt16)min,(UInt8*)&buf[length+1]);
		}
		length += 3;
	}
	if(buf) buf[length] = -2;	/* end of arriveLine */
	length++;
	if(trlen){
		i = length - trpos;
		if(buf) storeWord((UInt16)i,(UInt8*)&buf[trpos]);
		for (i=0; i < trlen; i++){
	    if(buf) buf[length] = trans[i];
	    length++;
		}
		if(buf) buf[length] = '\0';
		length++;
	}
	if(*p == linefeedChr) p++;
	*pp = p;
	return length;
}
 
static void passDB(Int32 recordNum, TimeTableType *table,
									 Int16 *tlenp, Int16 *clenp, Int16 *alenp,
									 CommentMarkType *commentsMark){
	//static Int16 tcount;
	static Int16 tcount;
	Int16 count;
	Char *p;
	Int16 tlen, clen, alen;
	Boolean ok = true;
	TimeTableGlobals_t* gTTP = NULL;
	gTTP = getPtrTimeTableGlobals();

	//DEBUG_LINE(33);
	gTTP->ContFileNo = 0;
	gTTP->ContFileRecNo = recordNum;
	contNextChar(&p);
	skipSpace(&p);
	titleLine(&p);
	table->LinesLen = 0;
	table->ArrivesLen = 0;
	table->NarrowsLen = NarrowsOffset;
	gTTP->NarrowList[0] = -1;
	tlen = clen = alen = 0;
	count = 0;
	//DEBUG_LINE(-1);
	if(table->LinesBuf == NULL) tcount = count;	/* pass 1 */
	//DEBUG_LINE(-1);
	gTTP->CommentMark = Mark;	/* underline */
	while(contNextChar(&p)){
		count++;
		skipSpace(&p);
		if(*p == '['){
	    ok = weekTypeLine(&p);
	    if(ok && table->LinesBuf == NULL){	/* pass 1 */
				tcount = count;		/* reload from here in pass 2 */
				tlen = 0;
				table->LinesLen = 0;
	    }
	    continue;
		}
		if(!ok){
	    skipLine(&p);
	    continue;
		}
		if(*p == ';'){
	    if(*(p+1) == ';'){
				p += 2;
				directiveLine(&p);
	    }else{
				skipLine(&p);	/* comment line */
	    }
		}else if(commentChar(*p) >= 0){
	    clen += commentLine(&p,(table->CommentsBuf ?
															table->CommentsBuf + clen : NULL),
													table->Comments,commentsMark,
													table->Narrows,&table->NarrowsLen);
		}else if(*p == '#'){
	    titleLine(&p);	/* title */
		}else if('0' <= *p && *p <= '9'){
	    if(table->LinesBuf == NULL
				 || count >= tcount){	/* in pass 2, reload from tcount */
				tlen += hourLine(&p,(LineType*)(table->LinesBuf ?
																				table->LinesBuf + tlen : NULL),
												 table->Lines,&table->LinesLen,
												 commentsMark);
			}
	    else skipLine(&p);
		}else if(*p == '$'){
	    alen += arriveLine(&p,(table->ArrivesBuf ?
														 table->ArrivesBuf + alen : NULL),
												 table->Arrives,&table->ArrivesLen);
		}else{
	    skipLine(&p);
		}
	}
	*tlenp = tlen;
	*clenp = clen;
	*alenp = alen;
}

static PrivErrType convTrainTimeDB(Int32 recordNum, TimeTableType *table){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	CommentMarkType commentsMark[NComments];
	Int16 tlen, clen, alen;
	Int16 tlen2, clen2, alen2;
	Int16 i, j;

	//DEBUG_LINE(27);
	for (i=0; i < NComments; i++) commentsMark[i] = UnMark;
	for (i=0; i <= 0x1F; i++){	/* initialize DotPattern */
		gTTP->DotPattern[i] = 0;
		for (j=0x10; j > 0; j >>= 1){
	    gTTP->DotPattern[i] <<= 2;
	    if(i & j) gTTP->DotPattern[i] |= 0x3;
		}
		gTTP->DotPattern[i] <<= 2;
	}
	//DEBUG_LINE(-1);
	passDB(recordNum,table,&tlen,&clen,&alen,commentsMark);	/* pass 1 */
	//DEBUG_LINE(-1);
	if(gTTP->ContFileHandle){
		//PrivErrDisplay("Can't happen: ContFileHandle != NULL",true);
		MemHandleUnlock(gTTP->ContFileHandle);
		gTTP->ContFileHandle = NULL;
	}
	if(tlen > 0 && table->LinesLen > 0){
		if(table->LinesBuf) MemPtrFree(table->LinesBuf);
		table->LinesBuf = (UInt8*)MemPtrNew(tlen);
		if(!table->LinesBuf){
	    //gTTP->PrivErrExtraStr = "table->LinesBuf";
	    return privErrNotEnoughSpace;
		}
		if(table->Lines) MemPtrFree(table->Lines);
		table->Lines = (LineType**)MemPtrNew(sizeof(LineType*) * table->LinesLen);
		if(!table->Lines){
	    gTTP->PrivErrExtraStr = "table->Lines";
	    return privErrNotEnoughSpace;
		}
	}
	if(clen > 0){
		if(table->Comments) MemPtrFree(table->Comments);
		table->Comments = (Char**)MemPtrNew(sizeof(Char*) * NComments);
		if(!table->Comments){
	    gTTP->PrivErrExtraStr = "table->Comments";
	    return privErrNotEnoughSpace;
		}
		if(table->CommentsBuf) MemPtrFree(table->CommentsBuf);
		table->CommentsBuf = (Char*)MemPtrNew(clen);
		if(!table->CommentsBuf){
	    gTTP->PrivErrExtraStr = "table->CommentsBuf";
	    return privErrNotEnoughSpace;
		}
		for (i=0; i < NComments; i++) table->Comments[i] = "";
	}
	if(table->Narrows) MemPtrFree(table->Narrows);
	table->Narrows = (Char**)MemPtrNew(sizeof(Char*) * table->NarrowsLen);
	if(!table->Narrows){
		gTTP->PrivErrExtraStr = "table->Narrows";
		return privErrNotEnoughSpace;
	}
	if(alen > 0 && table->ArrivesLen > 0){
		if(table->ArrivesBuf) MemPtrFree(table->ArrivesBuf);
		table->ArrivesBuf = (Char*)MemPtrNew(alen);
		if(!table->ArrivesBuf){
	    gTTP->PrivErrExtraStr = "table->ArrivesBuf";
	    return privErrNotEnoughSpace;
		}
		if(table->Arrives) MemPtrFree(table->Arrives);
		table->Arrives = (Char**)MemPtrNew(sizeof(Char*) * table->ArrivesLen);
		if(!table->Arrives){
	    gTTP->PrivErrExtraStr = "table->Arrives";
	    return privErrNotEnoughSpace;
		}
	}
	passDB(recordNum,table,&tlen2,&clen2,&alen2,commentsMark);	/* pass 2 */
	if(gTTP->ContFileHandle){
		//PrivErrDisplay("Can't happen: ContFileHandle != NULL",true);
		MemHandleUnlock(gTTP->ContFileHandle);
		gTTP->ContFileHandle = NULL;
	}
	if(tlen2 != tlen){
		gTTP->PrivErrExtraStr = "tlen2 != tlen";
	}else if(clen2 != clen){
		gTTP->PrivErrExtraStr = "clen2 != clen";
	}else if(alen2 != alen){
		gTTP->PrivErrExtraStr = "alen2 != alen";
	}
	if(gTTP->PrivErrExtraStr) return privErrCanNotHappen;
	if(tlen <= 0 || table->LinesLen <= 0) table->LinesLen = 0;
	if(alen > 0 && table->ArrivesLen > 0){
		if(gTTP->ArriveTime) MemPtrFree(gTTP->ArriveTime);
		gTTP->ArriveTime = (Int16*)MemPtrNew(sizeof(Int16) * table->ArrivesLen * gTTP->ArrivesColumn);
		if(!gTTP->ArriveTime){
	    gTTP->PrivErrExtraStr = "ArriveTime";
	    return privErrNotEnoughSpace;
		}
		if(gTTP->OfsArrive >= table->ArrivesLen) gTTP->OfsArrive = 0;
	}else{
		table->ArrivesLen = 0;
	}
	return privErrNone;
}

/*static*/ Boolean loadDB(Int32 recordNum){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	TimeTableType *table;
	PrivErrType Err;

	setToday(false);
	table = newTimeTable();
	if(table){
		//byHAL//UInt32 diff = TimGetSeconds() - gTTP->TodaySetSeconds;
		UInt32 diff = TimDateTimeToSeconds(&(gTTP->gHWP->cDateTime)) - gTTP->TodaySetSeconds;
		gTTP->TodayDate = Errday;
		if(diff >= 10 * 60 /* 10 min */ && gTTP->ConfigHandle){
	    gTTP->TodayDate = parseConfig(gTTP->ConfigHandle);
		}
		Err = convTrainTimeDB(recordNum, table);
	}else{
		gTTP->PrivErrExtraStr = "newTimeTable";
		Err = privErrNotEnoughSpace;
	}
	if(Err){
		;//PrivErrDisplay(PrivErrString(Err),true);
	}else if(table->LinesLen > 0){
		if(gTTP->CurrentTable) disposeTimeTable(gTTP->CurrentTable);
		table->recordNum = recordNum;
		gTTP->CurrentTable = table;
		return true;
	}
	disposeTimeTable(table);
	return false;
}

Boolean checkTitle(Int32 recordNum, Char *title){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	MemHandle record;
	Char str[STRMAX];
	Char *p;
	Int16 i;
	if(recordNum >= TimeTableRecordNumMin && gTTP->TimeTableDB){
		record = DmQueryRecord(gTTP->TimeTableDB,recordNum - TimeTableRecordNumMin);
	}else{
		record = DmQueryRecord(gTTP->MemoDB,recordNum);
	}
	if(record == NULL) return false;
	p = MemHandleLock(record);
	gTTP->ContFileNo = 0;
	(void)isContFile(&p);
	for (i=0; i < STRMAX-1; i++){
		if(!*p || *p == linefeedChr) break;
		str[i] = *p++;
	}
	str[i] = '\0';
	MemHandleUnlock(record);
	return StrStr(str,title) != NULL;
}

Int32 searchFile(Char *p){
	TimeTableGlobals_t* gTTP = getPtrTimeTableGlobals();
	UInt16 recordNum;
	Int32 ret;
	MemHandle record;
	if(gTTP->TimeTableDB){
		recordNum = 0;
		while((record=DmQueryNextInCategory(gTTP->TimeTableDB,&recordNum,
																				gTTP->TimeTableCategory))){
	    ret = recordNum++;
	    if(checkTitle(ret+TimeTableRecordNumMin,p))
				return ret + TimeTableRecordNumMin;
		}
	}
	recordNum = 0;
	while((record=DmQueryNextInCategory(gTTP->MemoDB,&recordNum,
																			gTTP->TrainTimeCategory))){
		ret = recordNum++;
		if(checkTitle(ret,p)) return ret;
	}
	return -1;
}
